/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */



/* Queue type to model a broadband link
 * -------------------------------------
 * Written Jan 2007 by Marcel Dischinger, MPI-SWS
 * Uses code from Andreas Haeberlen, MPI-SWS
 */

#undef NDEBUG
#include <assert.h>
#include "broadbandqueue.h"

/* Internal event handler */
void BBQueueHandler::handle(Event*)
{
	queue_.handleNextEvent();
}

static class BBQueueClass : public TclClass {
 public:
	BBQueueClass() : TclClass("Queue/BBQueue") {}
	TclObject* create(int, const char*const*) {
		return (new BBQueue);
	}
} class_drop_tail;


void BBQueue::reset()
{
	// From queue.cc
	total_time_ = 0.0;
	true_ave_ = 0.0;

	// Clear queue
	Packet* p;
	while ((p = q_->deque()) != 0)
		drop(p);

	// Clear internal events
	Scheduler& s = Scheduler::instance();
	if(intr_.uid_ > 0)
		s.cancel(&intr_);

	// Clear BroadbandQueue
	bq_.reset();

}


int 
BBQueue::command(int argc, const char*const* argv) 
{
	return Queue::command(argc, argv);
}


/* This is called every time this queue receives a new packet.
 * All we do here is 
 *   - append it to the BroadbandQueue
 *   - store it in a ns2 FIFO queue
 *   - schedule an event to send the next packet (if needed)
 */
void BBQueue::enque(Packet* p)
{
	Scheduler& s = Scheduler::instance();
	int size = hdr_cmn::access(p)->size();
	int id = p->uid_;
	long long now = getTimeMicros(s.clock());

	// Append the packet to the model
	bool appended = bq_.append(id, size, now);

	if(! appended)
		drop(p);
	else
		q_->enque(p);

	// Calculate if and when the next event should take place
	long long nextEventMillis = bq_.nextEvent();

	if(nextEventMillis != -1){
		double nextEventSecs = getTimeSecs(nextEventMillis);

		// If there is a pending event, reschedule it
		if((intr_.uid_ > 0) && (intr_.time_ != nextEventSecs)){
			s.cancel(&intr_);

			nextEventSecs = nextEventSecs - s.clock();
			s.schedule(&bh_, &intr_, (nextEventSecs >= 0) ? nextEventSecs : 0);
		}
	}

}

/* This is called every time the link can receive a new packet. 
 * If there's already an internal event pending, we do nothing.
 * Otherwise, we check when the next packet should be send and either send one 
 * or schedule an event for the next packet that should be sent.
 */
Packet* BBQueue::deque()
{
	link_is_ready_ = true;

	// We already have scheduled an event or there are no packets left in the queue
	if((intr_.uid_ > 0) || (q_->length() <= 0))
		return 0;
	else{
		Scheduler& s = Scheduler::instance();
		long long now = getTimeMicros(s.clock());
		
		// Check if we need to send a packet immediately
		long long nextEventMillis = bq_.nextEvent();

		if(nextEventMillis == -1){
			printf("BUG: deque() called and no event pending or packet to process\n");
			abort();
		}

		// Next event is due, return next packet
		if(nextEventMillis <= now){
			link_is_ready_ = false;

			int id = bq_.getNextPacket(now);
//			assert(id >= 0);

			Packet *p = q_->head();

			if(p->uid_ != id){
				printf("BUG: Next packet in queue has id %d while I expected id %d.\n", (int) p->uid_, id);
				abort();
			}

			return q_->deque();
		}

		// Schedule next event
		double nextEventSecs = getTimeSecs(nextEventMillis);
		if((nextEventSecs - s.clock()) <= 0)
			nextEventSecs = 0;
		else
			nextEventSecs = nextEventSecs - s.clock();
		
		s.schedule(&bh_, &intr_, nextEventSecs);
	}

	return 0;
}

/* This is called every time our internal event is scheduled.
 * If the link is ready to receive a packet, we send the next packet and
 * schedule a new event for the next packet.
 */
void BBQueue::handleNextEvent()
{
	Scheduler& s = Scheduler::instance();
	long long now = getTimeMicros(s.clock());

	if(link_is_ready_){
		link_is_ready_ = false;

		int id = bq_.getNextPacket(now);
		//		assert(id >= 0);
		assert(q_->length() > 0);

		Packet *p = q_->head();

		if(p->uid_ != id){
			printf("BUG: Next packet in queue has id %d while I expected id %d (in queue: %d, now=%lld, next=%lld).\n", (int) p->uid_, id, q_->length(), now, bq_.nextEvent());
			abort();

#if 0
			/* This would support reordering. But then we need to return p and not q_->deque() below
			 * Adjustments would be necessary in all the functions above! However, BroadbandQueue does
			 * not reorder packets.
			 */

			p = p->lookup(id);
			if(p == 0){
				printf("BUG: Packet with id %d not found!\n", id);
				abort();
			}
			q_->remove(p);
#endif
		}


		// Calculate when the next event should take place
		long long nextEventMillis = bq_.nextEvent();

		// Schedule next event
		if(nextEventMillis != -1){
			double nextEventSecs = getTimeSecs(nextEventMillis);
			if((nextEventSecs - s.clock()) <= 0)
				nextEventSecs = 0;
			else
				nextEventSecs = nextEventSecs - s.clock();

			s.schedule(&bh_, &intr_, nextEventSecs);
		}		

		send(q_->deque(), &qh_); // target_->recv(p, &qh_);
	}
}
